import json
from pathlib import Path

import click
import pandas

from sdgx.models.manager import ModelManager


@click.command()
@click.option(
    "--model",
    help="Name of model, use `sdgx list-models` to list all available models.",
    required=True,
)
@click.option(
    "--model_params",
    default="{}",
    help="[Json-string] Parameters for model.",
)
@click.option(
    "--input_path",
    help="Path of input data.",
    required=True,
)
@click.option(
    "--input_type",
    default="csv",
    help="Type of input data, will be used as `pandas.read_{input_type}`.",
)
@click.option(
    "--read_params",
    default="{}",
    help="[Json-string] Parameters for `pandas.read_{input_type}`.",
)
@click.option(
    "--fit_params",
    default="{}",
    help="[Json-string] Parameters for `model.fit`.",
)
@click.option(
    "--output_path",
    help="Path to save the model.",
    required=True,
)
def fit(
    model,
    model_params,
    input_path,
    input_type,
    read_params,
    fit_params,
    output_path,
):
    model_params = json.loads(model_params)
    read_params = json.loads(read_params)
    fit_params = json.loads(fit_params)

    model = ModelManager().init_model(model, **model_params)
    input_method = getattr(pandas, f"read_{input_type}")
    if not input_method:
        raise NotImplementedError(f"Pandas not support read_{input_type}")
    df = input_method(input_path, **read_params)
    model.fit(df, **fit_params)

    Path(output_path).mkdir(parents=True, exist_ok=True)
    model.save(output_path)


@click.command()
def sample(
    model_path,
    output_path,
    write_type,
    write_param,
):
    model = ModelManager.load(model_path)
    # TODO: Model not have `sample` in Base Class yet
    # sampled_data = model.sample()
    # write_method = getattr(sampled_data, f"to_{write_type}")
    # write_method(output_path, write_param)


@click.command()
def list_models():
    for model_name, model_cls in ModelManager().registed_model.items():
        print(f"{model_name} is registed as class: {model_cls}.")


@click.group()
def cli():
    pass


cli.add_command(fit)
cli.add_command(sample)
cli.add_command(list_models)



import contextlib
import logging
import threading

import numpy as np
import torch


# 添加增加的日志log
def get_log_file_handler(log_path):
    # 具体参数请参考：
    # https://docs.python.org/3/library/logging.handlers.html#timedrotatingfilehandler
    # 查路径请参考：
    # fileshandle.baseFilename
    fileshandler = logging.handlers.TimedRotatingFileHandler(
        log_path, when="W6", interval=5, backupCount=15, encoding="utf-8"
    )
    fileshandler.suffix = "%Y%m%d_%H%M%S.log"
    fileshandler.setLevel(logging.DEBUG)
    fmt_str = "%(asctime)s %(levelname)s %(filename)s[%(lineno)d] %(message)s"
    formatter = logging.Formatter(fmt_str)
    fileshandler.setFormatter(formatter)
    return fileshandler


# 引入自 ctgan
# 未来会根据业务需求优化修改
def random_state(function):
    """Set the random state before calling the function.

    Args:
        function (Callable):
            The function to wrap around.
    """

    def wrapper(self, *args, **kwargs):
        if self.random_states is None:
            return function(self, *args, **kwargs)

        else:
            with set_random_states(self.random_states, self.set_random_state):
                return function(self, *args, **kwargs)

    return wrapper


@contextlib.contextmanager
def set_random_states(random_state, set_model_random_state):
    """Context manager for managing the random state.

    Args:
        random_state (int or tuple):
            The random seed or a tuple of (numpy.random.RandomState, torch.Generator).
        set_model_random_state (function):
            Function to set the random state on the model.
    """
    original_np_state = np.random.get_state()
    original_torch_state = torch.get_rng_state()

    random_np_state, random_torch_state = random_state

    np.random.set_state(random_np_state.get_state())
    torch.set_rng_state(random_torch_state.get_state())

    try:
        yield
    finally:
        current_np_state = np.random.RandomState()
        current_np_state.set_state(np.random.get_state())
        current_torch_state = torch.Generator()
        current_torch_state.set_state(torch.get_rng_state())
        set_model_random_state((current_np_state, current_torch_state))

        np.random.set_state(original_np_state)
        torch.set_rng_state(original_torch_state)


def flatten_array(nested, prefix=""):
    """Flatten an array as a dict.

    Args:
        nested (list, numpy.array):
            Iterable to flatten.
        prefix (str):
            Name to append to the array indices. Defaults to ``''``.

    Returns:
        dict:
            Flattened array.
    """
    result = {}
    for index in range(len(nested)):
        prefix_key = "__".join([prefix, str(index)]) if len(prefix) else str(index)

        value = nested[index]
        if isinstance(value, (list, np.ndarray)):
            result.update(flatten_array(value, prefix=prefix_key))

        elif isinstance(value, dict):
            result.update(flatten_dict(value, prefix=prefix_key))

        else:
            result[prefix_key] = value

    return result


IGNORED_DICT_KEYS = ["fitted", "distribution", "type"]


def flatten_dict(nested, prefix=""):
    """Flatten a dictionary.

    This method returns a flatten version of a dictionary, concatenating key names with
    double underscores.

    Args:
        nested (dict):
            Original dictionary to flatten.
        prefix (str):
            Prefix to append to key name. Defaults to ``''``.

    Returns:
        dict:
            Flattened dictionary.
    """
    result = {}

    for key, value in nested.items():
        prefix_key = "__".join([prefix, str(key)]) if len(prefix) else key

        if key in IGNORED_DICT_KEYS and not isinstance(value, (dict, list)):
            continue

        elif isinstance(value, dict):
            result.update(flatten_dict(value, prefix_key))

        elif isinstance(value, (np.ndarray, list)):
            result.update(flatten_array(value, prefix_key))

        else:
            result[prefix_key] = value

    return result


def log_numerical_distributions_error(numerical_distributions, processed_data_columns, logger):
    """Log error when numerical distributions columns don't exist anymore."""
    unseen_columns = numerical_distributions.keys() - set(processed_data_columns)
    for column in unseen_columns:
        logger.info(
            f"Requested distribution '{numerical_distributions[column]}' "
            f"cannot be applied to column '{column}' because it no longer "
            "exists after preprocessing."
        )


def _key_order(key_value):
    parts = []
    for part in key_value[0].split("__"):
        if part.isdigit():
            part = int(part)

        parts.append(part)

    return parts


def unflatten_dict(flat):
    """Transform a flattened dict into its original form.

    Args:
        flat (dict):
            Flattened dict.

    Returns:
        dict:
            Nested dict (if corresponds)
    """
    unflattened = {}

    for key, value in sorted(flat.items(), key=_key_order):
        if "__" in key:
            key, subkey = key.split("__", 1)
            subkey, name = subkey.rsplit("__", 1)

            if name.isdigit():
                column_index = int(name)
                row_index = int(subkey)

                array = unflattened.setdefault(key, [])

                if len(array) == row_index:
                    row = []
                    array.append(row)
                elif len(array) == row_index + 1:
                    row = array[row_index]
                else:
                    # This should never happen
                    raise ValueError("There was an error unflattening the extension.")

                if len(row) == column_index:
                    row.append(value)
                else:
                    # This should never happen
                    raise ValueError("There was an error unflattening the extension.")

            else:
                subdict = unflattened.setdefault(key, {})
                if subkey.isdigit():
                    subkey = int(subkey)

                inner = subdict.setdefault(subkey, {})
                inner[name] = value

        else:
            unflattened[key] = value

    return unflattened


def validate_numerical_distributions(numerical_distributions, metadata_columns):
    """Validate ``numerical_distributions``.

    Raise an error if it's not None or dict, or if its columns are not present in the metadata.

    Args:
        numerical_distributions (dict):
            Dictionary that maps field names from the table that is being modeled with
            the distribution that needs to be used.
        metadata_columns (list):
            Columns present in the metadata.
    """
    if numerical_distributions:
        if not isinstance(numerical_distributions, dict):
            raise TypeError("numerical_distributions can only be None or a dict instance.")

        invalid_columns = numerical_distributions.keys() - set(metadata_columns)
        # if invalid_columns:
        #     raise SynthesizerInputError(
        #         'Invalid column names found in the numerical_distributions dictionary '
        #         f'{invalid_columns}. The column names you provide must be present '
        #         'in the metadata.'
        #     )


class Singleton(type):
    _instances = {}
    _lock = threading.Lock()

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            with cls._lock:
                if cls not in cls._instances:
                    cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]



from abc import ABC, abstractmethod


class BaseMetric(ABC):
    def __init__(self, real_data, synthetic_data):
        self.real_data = real_data
        self.synthetic_data = synthetic_data

    @abstractmethod
    def calculate(self):
        pass

    def validate_datasets(self):
        pass



# 本例用于测试 transformer 对象的正常使用

import os
import shutil
import sys

import pytest

_HERE = os.path.dirname(__file__)
sys.path.append(os.getcwd())

from sdgx.utils.io.csv_utils import *


def test_csv_tools():
    pass


if __name__ == "__main__":
    pytest.main(["-vv", "-s", __file__])



import os

import pytest

_HERE = os.path.dirname(__file__)

import sys

sys.path.append(os.getcwd())


from sdgx.models.single_table.ctgan import CTGAN
from sdgx.utils.io.csv_utils import *


def test_helloworld():
    # 读取数据
    demo_data, discrete_cols = get_demo_single_table()
    model = CTGAN(epochs=1)
    model.fit(demo_data, discrete_cols)

    # 生成合成数据
    sampled_data = model.sample(1000)


if __name__ == "__main__":
    pytest.main(["-vv", "-s", __file__])



from __future__ import annotations

from sdgx.models.base import BaseSynthesizerModel


class MyOwnModel(BaseSynthesizerModel):
    ...


from sdgx.models.extension import hookimpl


@hookimpl
def register(manager):
    manager.register("DummyModel", MyOwnModel)



from __future__ import annotations

from typing import Any

import pluggy

project_name = "sdgx.model"
hookimpl = pluggy.HookimplMarker(project_name)
hookspec = pluggy.HookspecMarker(project_name)


@hookspec
def register(manager):
    """
    For more information about this function, please check the ``ModelManager``

    We provided an example package for you in {project_root}/example/extension/dummymodel.

    Example:
    .. code-block:: python

        class MyOwnModel(BaseSynthesizerModel):
            ...

        from sdgx.models.extension import hookimpl

        @hookimpl
        def register(manager):
            manager.register("DummyModel", MyOwnModel)


    Config ``project.entry-points`` so that we can find it

    .. code-block:: toml
    [project.entry-points."sdgx.model"]
    < whatever-name > = "<package>.<path>.<to>.<model-file>"


    You can verify it by `sdgx list-models`.
    """



import os

import pymysql
from config import sdg_log_dir
from pymysql.err import IntegrityError, OperationalError

# logger 需要解决
from sdgx.utils import logger
from sdgx.utils.utils import get_log_file_handler


class mysql_db_connector:
    def __init__(
        self,
        databaes_ip,
        database_port,
        database_username,
        database_passwd,
        database_name,
        database_type="mysql",
    ) -> None:
        # 配置log路径
        if not os.path.exists(sdg_log_dir):
            os.makedirs(sdg_log_dir)
        log_path = os.path.join(sdg_log_dir, "mysql_log.log")
        log_path_handler = get_log_file_handler(log_path)
        logger.addHandler(log_path_handler)

        # 进行初始化
        self.database_type = database_type
        self.databaes_ip = databaes_ip
        self.database_port = database_port
        self.database_username = database_username
        self.database_passwd = database_passwd
        self.database_name = database_name

        # 记录状态
        self.conn = None
        self.status = "waiting"
        self.insert_success = False
        self.del_success = False

        # 建立一个 conn，
        if self.database_type == "mysql":
            logger.info(
                "Establish Mysql Connection with %s:%s" % (self.databaes_ip, self.database_port)
            )
            try:
                self.conn = pymysql.connect(
                    host=self.databaes_ip,
                    port=self.database_port,
                    user=self.database_username,
                    password=self.database_passwd,
                    database=self.database_name,
                )
                self.status = "success"
                logger.info("MySQL Connection Created.")
            except ConnectionRefusedError:
                self.status = "failed"
                logger.error("MySQL Connection Refused")
            except OperationalError:
                logger.error("MySQL OperationalError")
                self.status = "failed"
        else:
            logger.error("Not support other Database yet.")
            raise NotImplementedError("Not support other Database yet.")
        pass

    def query(self, sql):
        # 查询动作，把数据给提出来，以列表形式返回
        # assert 'select' in sql.lowercase()
        if "select" not in sql.lower():
            raise ValueError("NOT a Query SQL")
        cursor = self.conn.cursor()
        select_cnt = cursor.execute(sql)  # 得到结果行数
        res = []
        for _ in range(select_cnt):
            sample = list(cursor.fetchone())
            res.append(sample)
        cursor.close()
        logger.debug('Query: "%s", Success' % sql.replace("\n", " ").strip())
        return res  # 结果一定是二维数组

    def insert(self, sql):
        # INSERT INTO `fed_sql`.`b_job` (`id`, `job_id`, `job_sql`, `job_ priority`, `grammar_check`, `integrity_check`, `status`, `create_time`, `update_time`, `start_time`) VALUES ('49', '1', 'selec * from A.xx;', '2', '1', '1', 'success', '1657871901', '1657871901', '1657871901');
        if "insert" not in sql.lower() and "update" not in sql.lower():
            raise ValueError("NOT a Insert SQL")
        cursor = self.conn.cursor()
        try:
            cursor.execute(sql)
            cursor.close()
        except IntegrityError:
            logger.error('Insert: "%s", Failed' % sql.replace("\n", " "))
            self.insert_success = False
        cursor.close()
        self.conn.commit()
        if "insert" in sql.lower():
            logger.debug('Insert: "%s", Success' % sql.replace("\n", " "))
        elif "update" in sql.lower():
            logger.debug('Update: "%s", Success' % sql.replace("\n", " "))
        self.insert_success = True

    def delete(self, sql):
        # DELETE FROM `fed_sql`.`s_party` WHERE (`id` = '3');
        if "delete" not in sql.lower():
            raise ValueError("NOT a Delete SQL")
        cursor = self.conn.cursor()
        try:
            cursor.execute(sql)
            cursor.close()
        except Exception as e:
            logger.error('Delete: "%s", Failed' % sql.replace("\n", " "))
            logger.error(str(e))
            self.del_success = False
        cursor.close()
        self.conn.commit()
        logger.debug('Delete: "%s", Success' % sql.replace("\n", " "))
        self.del_success = True

    def get_tables(self):
        sql = "show tables;"
        cursor = self.conn.cursor()
        table_cnt = cursor.execute(sql)
        res = []
        for _ in range(table_cnt):
            table = cursor.fetchone()[0]
            res.append(table)
        return res

    def create_table(self, sql):
        if "create" not in sql.lower():
            raise ValueError("NOT a Create SQL")
        cursor = self.conn.cursor()
        cursor.execute(sql)
        self.conn.commit()
        cursor.close()

    def __del__(self):
        # 关闭数据库连接
        if self.status == "failed":
            logger.info("NO MySQL Connection, Closed.")
        elif self.status != "closed":
            self.conn.close()
            logger.info("MySQL Connection Closed.")

    # 手工 close
    # 一般不太用得到
    def close(self):
        self.conn.close()
        self.status = "closed"
        logger.info("MySQL Connection Closed in func <close>.")



from __future__ import annotations

import glob
import importlib
from os.path import basename, dirname, isfile, join
from typing import Any

import pluggy

from sdgx import models
from sdgx.errors import ModelInitializationError, ModelNotFoundError, ModelRegisterError
from sdgx.log import logger
from sdgx.models import extension
from sdgx.models.base import BaseSynthesizerModel
from sdgx.models.extension import project_name as PROJECT_NAME
from sdgx.utils.utils import Singleton


class ModelManager(metaclass=Singleton):
    def __init__(self):
        self.pm = pluggy.PluginManager(PROJECT_NAME)
        self.pm.add_hookspecs(extension)
        self._registed_model: dict[str, type[BaseSynthesizerModel]] = {}

        # Load all
        self.pm.load_setuptools_entrypoints(PROJECT_NAME)

        # Load all local model
        self.load_all_local_model()

    @property
    def registed_model(self):
        # Lazy load when query registed_model
        if self._registed_model:
            return self._registed_model
        for f in self.pm.hook.register(manager=self):
            try:
                f()
            except Exception as e:
                logger.exception(ModelRegisterError(e))
                continue
        return self._registed_model

    def load_all_local_model(self):
        # self.pm.register(sdgx/models/single_table/*)
        self._load_dir(models.single_table)
        # self.pm.register(sdgx/models/multi_tables/*)
        self._load_dir(models.multi_tables)

    def _load_dir(self, module):
        modules = glob.glob(join(dirname(module.__file__), "*.py"))
        sub_packages = (
            basename(f)[:-3] for f in modules if isfile(f) and not f.endswith("__init__.py")
        )
        packages = (str(module.__package__) + "." + i for i in sub_packages)
        for p in packages:
            self.pm.register(importlib.import_module(p))

    def _normalize_name(self, model_name: str) -> str:
        return model_name.strip().lower()

    def register(self, model_name, model_cls: type[BaseSynthesizerModel]):
        model_name = self._normalize_name(model_name)
        logger.info(f"Register for new model: {model_name}")
        self._registed_model[model_name] = model_cls

    def init_model(self, model_name, **kwargs: dict[str, Any]) -> BaseSynthesizerModel:
        model_name = self._normalize_name(model_name)
        if not model_name in self.registed_model:
            raise ModelNotFoundError
        try:
            return self.registed_model[model_name](**kwargs)
        except Exception as e:
            raise ModelInitializationError(e)

    @staticmethod
    def load(model_path) -> BaseSynthesizerModel:
        return BaseSynthesizerModel.load(model_path)



import os
from collections import namedtuple
from io import StringIO
from typing import List, Optional

import numpy as np
import pandas as pd
from joblib import Parallel, delayed
from rdt.transformers import ClusterBasedNormalizer, OneHotEncoder

from sdgx.utils.io.csv_utils import get_csv_column, get_csv_column_count

SpanInfo = namedtuple("SpanInfo", ["dim", "activation_fn"])
ColumnTransformInfo = namedtuple(
    "ColumnTransformInfo",
    ["column_name", "column_type", "transform", "output_info", "output_dimensions"],
)


# OPTIMIZE SDG 重写的 Data Transformer
class DataTransformer(object):
    """OPTIMIZE SDG 重写的 Data Transformer

    应对大数据（数据 > 内存）情况下的 Transformer 解决方案
        - 对于连续列：使用 BayesianGMM 对连续列建模并标准化为标量 [0, 1] 和向量。
        - 对于离散列：使用 scikit-learn OneHotEncoder 进行编码。

    本例的实现主要是为了支持大数据情况，并支持将 transform 的数据结果 以及 transformer
    所需的各类元数据 dump 到硬盘中，以实现较少的内存消耗。
    """

    def __init__(self, max_clusters=10, weight_threshold=0.005):
        """Create a data transformer.

        输入参数:
            max_clusters (int):
                Maximum number of Gaussian distributions in Bayesian GMM.
            weight_threshold (float):
                Weight threshold for a Gaussian distribution to be kept.
        """
        self._max_clusters = max_clusters
        self._weight_threshold = weight_threshold

    def _fit_continuous(self, column_name):
        """针对连续特征，训练一个贝叶斯 GMM， 用于 transform。
            该方法暂时不需要做进一步修改

        输入参数:
            data (pd.DataFrame):
                一个 pd.DataFrame 对象，但是只含有单列数据

        返回对象说明:
            namedtuple对象:
                返回单个 ``ColumnTransformInfo`` 对象
        """
        data = get_csv_column(self.raw_data_path, column_name)

        # column_name = data.columns[0]

        gm = ClusterBasedNormalizer(
            missing_value_generation="from_column", max_clusters=min(len(data), 10)
        )
        gm.fit(data, column_name)
        num_components = sum(gm.valid_component_indicator)

        # 使用 transform 对象存储单个列的 transformer
        return ColumnTransformInfo(
            column_name=column_name,
            column_type="continuous",
            transform=gm,
            output_info=[SpanInfo(1, "tanh"), SpanInfo(num_components, "softmax")],
            output_dimensions=1 + num_components,
        )

    def _fit_discrete(self, column_name):
        """针对离散特征：使用一个 one hot encoder 进行编码.
            该方法暂时不需要做进一步修改

        输入参数:
            data (pd.DataFrame):
                一个 pd.DataFrame 对象，但是只含有单列数据
            column_name (str)：
                需要进行 fit 操作的列名

        返回对象说明:
            namedtuple 对象:
                返回单个 ``ColumnTransformInfo`` 对象
        """
        data = get_csv_column(self.raw_data_path, column_name)
        # column_name = data.columns[0]
        ohe = OneHotEncoder()
        ohe.fit(data, column_name)
        num_categories = len(ohe.dummies)

        # 使用 transform 对象存储单个列的 transformer
        return ColumnTransformInfo(
            column_name=column_name,
            column_type="discrete",
            transform=ohe,
            output_info=[SpanInfo(num_categories, "softmax")],
            output_dimensions=num_categories,
        )

    # `fit` 是 transform 的入口方法
    def fit(self, raw_data_path, discrete_columns: Optional[List] = None, csv_cache_N=1000):
        """进行 DataTransformer 的 Fit 操作

        针对连续变量，使用 ``ClusterBasedNormalizer`` 进行 transform；
        对于离散变量，使用 ``OneHotEncoder``  进行 transform。

        此步骤还对矩阵数据和跨度信息中的#columns 进行计数操作。

        输入参数:
            raw_data_path (str):
                目前仅接受 csv 文件作为输入。
            discrete_columns (list):
                离散列名称，需要与 csv 文件的列名对应名，
                默认该参数为 [] 。
            csv_cache_N (int):
                为了减少内存，使用缓存 csv 的行数。
        """
        # raw data 需要从 path 中获取
        # raw_data = None
        self.raw_data_path = raw_data_path
        self.output_info_list = []
        self.output_dimensions = 0
        self.csv_cache_N = csv_cache_N
        # self.dataframe = False
        if not discrete_columns:
            discrete_columns = []

        # 对 data frame 的检测，如果不是 data frame ，还是转换为了 dataframe 进行操作
        # 但是经过搜索 此处仅对 后续一个逻辑又变化，所以我们直接把 self. dataframe 设置为 False
        # 在未来处理的时候，单独载入 data frame
        # if not isinstance(raw_data, pd.DataFrame):
        #     self. dataframe = False
        #     # work around for RDT issue #328 Fitting with numerical column names fails
        #     discrete_columns = [str(column) for column in discrete_columns]
        #     column_names = [str(num) for num in range(raw_data.shape[1])]
        #     raw_data = pd.DataFrame(raw_data, columns=column_names)

        # 对 discrete_columns 的处理还是保留了下来
        discrete_columns = [str(column) for column in discrete_columns]

        # 对 raw_data_path 增加检查
        #   - 是否存在，使用 os.path.exists 判断
        #   - 是否是 csv ，目前仅做名称上的检查
        if not os.path.exists(raw_data_path):
            raise FileNotFoundError("Input csv NOT Found.")
        if not "csv" in raw_data_path:
            raise ValueError("Input Data is NOT CSV.")

        # raw data columns 在原有的实现方法中使用了 raw_data.columns
        # 这么做的问题就在于 df 直接载入了所有数据
        # 为了减少内存消耗，我们通过读取 csv 的前 x 行，并借助原有代码完成相应功能的计算
        cache_csv_str = ""
        with open(raw_data_path, "r") as f:
            for _ in range(self.csv_cache_N + 1):
                cache_csv_str += f.readline()
        # 从 str 创建 data frame
        cache_io = StringIO(cache_csv_str)
        raw_data_cache = pd.read_csv(cache_io)

        # get columns
        self.column_names = [str(num) for num in range(raw_data_cache.shape[1])]

        self._column_raw_dtypes = raw_data_cache.infer_objects().dtypes
        self._column_transform_info_list = []

        for column_name in raw_data_cache.columns:
            if column_name in discrete_columns:
                column_transform_info = self._fit_discrete(column_name)
            else:
                column_transform_info = self._fit_continuous(column_name)

            self.output_info_list.append(column_transform_info.output_info)
            self.output_dimensions += column_transform_info.output_dimensions
            self._column_transform_info_list.append(column_transform_info)

    def _transform_continuous(self, column_transform_info, data):
        column_name = data.columns[0]
        flattened_column = data[column_name].to_numpy().flatten()
        data = data.assign(**{column_name: flattened_column})
        gm = column_transform_info.transform
        transformed = gm.transform(data)

        #  Converts the transformed data to the appropriate output format.
        #  The first column (ending in '.normalized') stays the same,
        #  but the lable encoded column (ending in '.component') is one hot encoded.
        output = np.zeros((len(transformed), column_transform_info.output_dimensions))
        output[:, 0] = transformed[f"{column_name}.normalized"].to_numpy()
        index = transformed[f"{column_name}.component"].to_numpy().astype(int)
        output[np.arange(index.size), index + 1] = 1.0

        return output

    def _transform_discrete(self, column_transform_info, data):
        ohe = column_transform_info.transform
        return ohe.transform(data).to_numpy()

    def _synchronous_transform(self, input_data_path, column_transform_info_list, output_path):
        """Take a csv path, and transform columns synchronous.

        Outputs a list with Numpy arrays to disk.
        """

        loop = True
        # has_write_header = True
        # use iterator = True
        reader = pd.read_csv(input_data_path, iterator=True, chunksize=1000000)
        while loop:
            column_data_list = []
            # get raw data
            raw_data = None
            try:
                raw_data = reader.get_chunk()
            except StopIteration:
                loop = False

            # break if loop is iteration is end
            if not loop:
                break

            """
            # write csv header to file
            if has_write_header:
                f = open(output_path, "w")
                f.write(",".join(raw_data.columns.to_list()) + '\n')
                f.close()
                has_write_header = False
            """

            # transform data here
            for column_transform_info in column_transform_info_list:
                column_name = column_transform_info.column_name
                # 获取列数据
                data = raw_data[[column_name]]
                if column_transform_info.column_type == "continuous":
                    column_data_list.append(self._transform_continuous(column_transform_info, data))
                else:
                    column_data_list.append(self._transform_discrete(column_transform_info, data))

            # 追加写到 output path 吧
            chunk_array = np.concatenate(column_data_list, axis=1).astype(float)
            # file object
            f = open(output_path, "a")
            np.savetxt(f, chunk_array, fmt="%g", delimiter=",")
            f.close()
        # end while
        # return column_data_list

    def _parallel_transform(self, raw_data, column_transform_info_list):
        """Take a Pandas DataFrame and transform columns in parallel.

        Outputs a list with Numpy arrays.
        """
        processes = []
        for column_transform_info in column_transform_info_list:
            column_name = column_transform_info.column_name
            data = raw_data[[column_name]]
            process = None
            if column_transform_info.column_type == "continuous":
                process = delayed(self._transform_continuous)(column_transform_info, data)
            else:
                process = delayed(self._transform_discrete)(column_transform_info, data)
            processes.append(process)

        return Parallel(n_jobs=-1)(processes)

    # TODO function to finish
    def transform(self, input_data_path, output_data_path):
        """Take raw data and output a matrix data."""
        # 从 disk 中读取
        raw_data = None

        #  check the parameter `input_data_path`
        if not os.path.exists(input_data_path):
            raise FileNotFoundError("Input csv NOT Found.")
        if not "csv" in input_data_path:
            raise ValueError("Input Data is NOT CSV.")

        # check the parameter `output data path`
        if not "csv" in output_data_path:
            raise ValueError("Input Data is NOT CSV.")

        # get the csv shape column count
        # import func from utils
        column_cnt = get_csv_column_count(input_data_path)

        # Only use parallelization with larger data sizes.
        # Otherwise, the transformation will be slower.
        # if column_cnt < 500:
        # column_data_list = self._synchronous_transform(
        self._synchronous_transform(
            input_data_path, self._column_transform_info_list, output_data_path
        )
        # else:
        #     column_data_list = self._parallel_transform(
        #         input_data_path,
        #         self._column_transform_info_list)

        # return np.concatenate(column_data_list, axis=1).astype(float)
        return output_data_path

    def _inverse_transform_continuous(self, column_transform_info, column_data, sigmas, st):
        gm = column_transform_info.transform
        data = pd.DataFrame(column_data[:, :2], columns=list(gm.get_output_sdtypes()))
        data[data.columns[1]] = np.argmax(column_data[:, 1:], axis=1)
        if sigmas is not None:
            selected_normalized_value = np.random.normal(data.iloc[:, 0], sigmas[st])
            data.iloc[:, 0] = selected_normalized_value

        return gm.reverse_transform(data)

    def _inverse_transform_discrete(self, column_transform_info, column_data):
        ohe = column_transform_info.transform
        data = pd.DataFrame(column_data, columns=list(ohe.get_output_sdtypes()))
        return ohe.reverse_transform(data)[column_transform_info.column_name]

    def inverse_transform(self, data, sigmas=None, output_path=None, write_header=True):
        """Take matrix data and output raw data.

        Output uses the same type as input to the transform function.
        Either np array or pd dataframe.
        """
        st = 0
        recovered_column_data_list = []
        column_names = []
        for column_transform_info in self._column_transform_info_list:
            dim = column_transform_info.output_dimensions
            column_data = data[:, st : st + dim]
            if column_transform_info.column_type == "continuous":
                recovered_column_data = self._inverse_transform_continuous(
                    column_transform_info, column_data, sigmas, st
                )
            else:
                recovered_column_data = self._inverse_transform_discrete(
                    column_transform_info, column_data
                )

            recovered_column_data_list.append(recovered_column_data)
            column_names.append(column_transform_info.column_name)
            st += dim

        recovered_data = np.column_stack(recovered_column_data_list)
        recovered_data = pd.DataFrame(recovered_data, columns=column_names).astype(
            self._column_raw_dtypes
        )

        # input is csv, self.dataframe will always be true
        # if not self.dataframe:
        #     recovered_data = recovered_data.to_numpy()
        if output_path is not None:
            f = open(output_path, "w")
            # write csv header to file
            if write_header:
                # f = open(output_path, "w")
                # recovered_data's type is pd.DataFrame
                f.write(",".join(recovered_data.columns.to_list()) + "\n")
            f.close()
            recovered_data.to_csv(output_path, index=False)

        return recovered_data

    def convert_column_name_value_to_id(self, column_name, value):
        """Get the ids of the given `column_name`."""
        discrete_counter = 0
        column_id = 0
        for column_transform_info in self._column_transform_info_list:
            if column_transform_info.column_name == column_name:
                break
            if column_transform_info.column_type == "discrete":
                discrete_counter += 1

            column_id += 1

        else:
            raise ValueError(f"The column_name `{column_name}` doesn't exist in the data.")

        ohe = column_transform_info.transform
        data = pd.DataFrame([value], columns=[column_transform_info.column_name])
        one_hot = ohe.transform(data).to_numpy()[0]
        if sum(one_hot) == 0:
            raise ValueError(f"The value `{value}` doesn't exist in the column `{column_name}`.")

        return {
            "discrete_column_id": discrete_counter,
            "column_id": column_id,
            "value_id": np.argmax(one_hot),
        }



# 运行该例子，可使用：
# ipython -i  example/2_guassian_copula_example.py
# 并查看 sampled_data 变量

from sdgx.statistics.single_table.copula import GaussianCopulaSynthesizer
from sdgx.utils.io.csv_utils import *

# 针对 csv 格式的小规模数据
# 目前我们以 df 作为输入的数据的格式
demo_data, discrete_cols = get_demo_single_table()
# print(demo_data)
# print(discrete_cols)

model = GaussianCopulaSynthesizer(discrete_cols)
model.fit(demo_data)

# sampled
sampled_data = model.sample(10)
print(sampled_data)



# 运行该例子，可使用：
# ipython -i  example/1_ctgan_example.py
# 并查看 sampled_data 变量

from sdgx.models.single_table.ctgan import CTGAN

# from sdgx.transform.sampler import DataSamplerCTGAN
# from sdgx.transform.transformer import DataTransformerCTGAN
from sdgx.utils.io.csv_utils import *

# 针对 csv 格式的小规模数据
# 目前我们以 df 作为输入的数据的格式
demo_data, discrete_cols = get_demo_single_table()

model = CTGAN(epochs=10)
model.fit(demo_data, discrete_cols)

# sampled
sampled_data = model.sample(1000)



import pytest

from sdgx.models.manager import ModelManager
from sdgx.utils.io.csv_utils import get_demo_single_table


@pytest.fixture
def model_manager():
    yield ModelManager()


def test_model_manager(model_manager: ModelManager):
    assert "ctgan" in model_manager.registed_model

    model = model_manager.init_model("ctgan", epochs=1)
    demo_data, discrete_cols = get_demo_single_table()
    model.fit(demo_data, discrete_cols)

    # 生成合成数据
    sampled_data = model.sample(1000)


if __name__ == "__main__":
    pytest.main(["-vv", "-s", __file__])



import os

# 多线程相关，目前不用
CPU_COUNT = os.cpu_count()  # 后续用于多线程
PORT = 7970  # 端口，目前未启用
SDG_PORT = PORT

# 日志路径
logger_config_path = "utilities/logging.conf"


# 路径配置，目前不太用
tmp_dir = "/tmp/sdg/tmp/"
sdg_data_dir = "/tmp/sdg/data/"
sdg_log_dir = "/tmp/sdg/log/"
# 分任务进行，目前不用
# sdg_job_dir   = "tmp/job/"



"""
    Wrappers around copulas models.
    需要修改: fit接口以适应性能优化措施
"""
import logging
import warnings
from copy import deepcopy

import copulas
import copulas.univariate
import numpy as np
import pandas as pd
import scipy
from copulas import multivariate
from rdt.transformers import OneHotEncoder

from sdgx.errors import NonParametricError
from sdgx.statistics.base import BaseSynthesizerModel

# transformer 以及 sampler 已经拆分，挪到 transform/ 目录中
# from sdgx.transform.sampler import DataSamplerCTGAN
from sdgx.transform.transformer import DataTransformerCTGAN
from sdgx.utils.utils import (
    flatten_dict,
    log_numerical_distributions_error,
    unflatten_dict,
    validate_numerical_distributions,
)

LOGGER = logging.getLogger(__name__)


class GaussianCopulaSynthesizer(BaseSynthesizerModel):
    """Model wrapping ``copulas.multivariate.GaussianMultivariate`` copula.

    Args:
        metadata (sdv.metadata.SingleTableMetadata):
            Single table metadata representing the data that this synthesizer will be used for.
        enforce_min_max_values (bool):
            Specify whether or not to clip the data returned by ``reverse_transform`` of
            the numerical transformer, ``FloatFormatter``, to the min and max values seen
            during ``fit``. Defaults to ``True``.
        enforce_rounding (bool):
            Define rounding scheme for ``numerical`` columns. If ``True``, the data returned
            by ``reverse_transform`` will be rounded as in the original data. Defaults to ``True``.
        locales (list or str):
            The default locale(s) to use for AnonymizedFaker transformers. Defaults to ``None``.
        numerical_distributions (dict):
            Dictionary that maps field names from the table that is being modeled with
            the distribution that needs to be used. The distributions can be passed as either
            a ``copulas.univariate`` instance or as one of the following values:

                * ``norm``: Use a norm distribution.
                * ``beta``: Use a Beta distribution.
                * ``truncnorm``: Use a truncnorm distribution.
                * ``uniform``: Use a uniform distribution.
                * ``gamma``: Use a Gamma distribution.
                * ``gaussian_kde``: Use a GaussianKDE distribution. This model is non-parametric,
                  so using this will make ``get_parameters`` unusable.

        default_distribution (str):
            Copulas univariate distribution to use by default. Valid options are:

                * ``norm``: Use a norm distribution.
                * ``beta``: Use a Beta distribution.
                * ``truncnorm``: Use a Truncated Gaussian distribution.
                * ``uniform``: Use a uniform distribution.
                * ``gamma``: Use a Gamma distribution.
                * ``gaussian_kde``: Use a GaussianKDE distribution. This model is non-parametric,
                  so using this will make ``get_parameters`` unusable.
             Defaults to ``beta``.
    """

    # 这里代表的是几种可选的分布，这里通常使用 beta
    # 其他分布的内容我们可以以后再了解。
    # 这个应该是属于 类变量的
    _DISTRIBUTIONS = {
        "norm": copulas.univariate.GaussianUnivariate,
        "beta": copulas.univariate.BetaUnivariate,
        "truncnorm": copulas.univariate.TruncatedGaussian,
        "gamma": copulas.univariate.GammaUnivariate,
        "uniform": copulas.univariate.UniformUnivariate,
        "gaussian_kde": copulas.univariate.GaussianKDE,
    }

    # model 变量就是承载模型的，使用 self._model 来访问
    _model = None

    # 这个方法很简单，就是把上面 _distribution 给 return 出来
    # 相当于一个字典的作用
    @classmethod
    def get_distribution_class(cls, distribution):
        """Return the corresponding distribution class from ``copulas.univariate``.

        Args:
            distribution (str):
                A string representing a copulas univariate distribution.

        Returns:
            copulas.univariate:
                A copulas univariate class that corresponds to the distribution.
        """
        if not isinstance(distribution, str) or distribution not in cls._DISTRIBUTIONS:
            error_message = f"Invalid distribution specification '{distribution}'."
            raise ValueError(error_message)

        return cls._DISTRIBUTIONS[distribution]

    # 初始化方法，参数的类型需要仔细分析一下
    def __init__(
        self,
        metadata,
        enforce_min_max_values=True,
        enforce_rounding=True,
        locales=None,
        numerical_distributions=None,
        default_distribution=None,
    ):
        self.metadata = (metadata,)  # fit 数据集的元数据
        self.enforce_min_max_values = (enforce_min_max_values,)  # 限制最大最小value，一般都是true
        self.enforce_rounding = (enforce_rounding,)  # 保持相同的小数点位数， 一般都是true
        self.locales = (locales,)  # 和语言设置有关，这个暂时不太清楚

        # 验证分布？
        # 这个时候还没有输入数据的，只有 metadata
        # 应该是通过 metadata 来验证分布的
        validate_numerical_distributions(numerical_distributions, self.metadata)
        # 这里的意思是，如果没有指定分布，那么就使用 beta 分布
        # 下面分别是两个参数，重新赋值
        self.numerical_distributions = numerical_distributions or {}
        self.default_distribution = default_distribution or "beta"

        # 以下几行代码是使用 class method 把字符串类型的分布
        # 转化为 copulas.univariate 的分布实例
        self._default_distribution = self.get_distribution_class(self.default_distribution)
        self._numerical_distributions = {
            field: self.get_distribution_class(distribution)
            for field, distribution in self.numerical_distributions.items()
        }
        # 最后是行数，目前是 none
        # 目前是没有训练数据的
        # 这个变量可能和训练数据有关
        self._num_rows = None

    # 训练函数，输入的是训练数据
    # 需要是 data frame 格式
    # 初步的性能优化方案可以是：
    # 1. 增加增量学习机制，防止内存消耗过多
    # 2. 增加
    def fit(self, processed_data):
        # 载入 transformer
        self._transformer = DataTransformerCTGAN()
        self._transformer.fit(processed_data, self.metadata[0])

        # 使用 transformer 处理数据
        processed_data = pd.DataFrame(self._transformer.transform(processed_data))

        # 这个应该是打 log ，不影响实际训练
        log_numerical_distributions_error(
            self.numerical_distributions, processed_data.columns, LOGGER
        )

        # 记录学习数据的长度
        self._num_rows = len(processed_data)

        # 这里是把没有指定分布的列，都设置为默认分布
        numerical_distributions = deepcopy(self._numerical_distributions)
        for column in processed_data.columns:
            if column not in numerical_distributions:
                numerical_distributions[column] = self._numerical_distributions.get(
                    column, self._default_distribution
                )

        # 然后建立模型
        # 模型直接调用的 multivariate.GaussianMultivariate
        # 这个GaussianMultivariate是可以直接 fit 的，只不过不在这里
        self._model = multivariate.GaussianMultivariate(distribution=numerical_distributions)

        # 在这里调用 fit 方法
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", module="scipy")
            self._model.fit(processed_data)
        # 注意看，这里只是忽略了警告，但是是直接调用的 data frame 用于 fit
        # 所以从内存角度而言，多大的数据都是直接载入到内存中的，优化需要从这里下手

    def _warn_for_update_transformers(self, column_name_to_transformer):
        """Raise warnings for update_transformers.

        Args:
            column_name_to_transformer (dict):
                Dict mapping column names to transformers to be used for that column.
        """
        for column, transformer in column_name_to_transformer.items():
            sdtype = self.metadata.columns[column]["sdtype"]
            if sdtype == "categorical" and isinstance(transformer, OneHotEncoder):
                warnings.warn(
                    f"Using a OneHotEncoder transformer for column '{column}' "
                    "may slow down the preprocessing and modeling times."
                )

    # 生成数据函数，这里其实直接调用的 model 的方法
    # 没有其他了

    def sample(self, num_rows, conditions=None):
        """Sample the indicated number of rows from the model.

        Args:
            num_rows (int):
                Amount of rows to sample.
            conditions (dict):
                If specified, this dictionary maps column names to the column
                value. Then, this method generates ``num_rows`` samples, all of
                which are conditioned on the given variables.

        Returns:
            pandas.DataFrame:
                Sampled data.
        """
        return self._transformer.inverse_transform(
            self._model.sample(num_rows, conditions=conditions).to_numpy()
        )

    def _get_valid_columns_from_metadata(self, columns):
        valid_columns = []
        for column in columns:
            for valid_column in self.metadata.columns:
                if column.startswith(valid_column):
                    valid_columns.append(column)
                    break

        return valid_columns

    def get_learned_distributions(self):
        """Get the marginal distributions used by the ``GaussianCopula``.

        Return a dictionary mapping the column names with the distribution name and the learned
        parameters for those.

        Returns:
            dict:
                Dictionary containing the distributions used or detected for each column and the
                learned parameters for those.
        """
        if not self._fitted:
            raise ValueError(
                "Distributions have not been learned yet. Please fit your model first using 'fit'."
            )

        parameters = self._model.to_dict()
        columns = parameters["columns"]
        univariates = deepcopy(parameters["univariates"])
        learned_distributions = {}
        valid_columns = self._get_valid_columns_from_metadata(columns)
        for column, learned_params in zip(columns, univariates):
            if column in valid_columns:
                distribution = self.numerical_distributions.get(column, self.default_distribution)
                learned_params.pop("type")
                learned_distributions[column] = {
                    "distribution": distribution,
                    "learned_parameters": learned_params,
                }

        return learned_distributions

    def _get_parameters(self):
        """Get copula model parameters.

        Compute model ``correlation`` and ``distribution.std``
        before it returns the flatten dict.

        Returns:
            dict:
                Copula parameters.

        Raises:
            NonParametricError:
                If a non-parametric distribution has been used.
        """
        for univariate in self._model.univariates:
            univariate_type = type(univariate)
            if univariate_type is copulas.univariate.Univariate:
                univariate = univariate._instance

            if univariate.PARAMETRIC == copulas.univariate.ParametricType.NON_PARAMETRIC:
                raise NonParametricError("This GaussianCopula uses non parametric distributions")

        params = self._model.to_dict()

        correlation = []
        for index, row in enumerate(params["correlation"][1:]):
            correlation.append(row[: index + 1])

        params["correlation"] = correlation
        params["univariates"] = dict(zip(params.pop("columns"), params["univariates"]))
        params["num_rows"] = self._num_rows

        return flatten_dict(params)

    @staticmethod
    def _get_nearest_correlation_matrix(matrix):
        """Find the nearest correlation matrix.

        If the given matrix is not Positive Semi-definite, which means
        that any of its eigenvalues is negative, find the nearest PSD matrix
        by setting the negative eigenvalues to 0 and rebuilding the matrix
        from the same eigenvectors and the modified eigenvalues.

        After this, the matrix will be PSD but may not have 1s in the diagonal,
        so the diagonal is replaced by 1s and then the PSD condition of the
        matrix is validated again, repeating the process until the built matrix
        contains 1s in all the diagonal and is PSD.

        After 10 iterations, the last step is skipped and the current PSD matrix
        is returned even if it does not have all 1s in the diagonal.

        Insipired by: https://stackoverflow.com/a/63131250
        """
        eigenvalues, eigenvectors = scipy.linalg.eigh(matrix)
        negative = eigenvalues < 0
        identity = np.identity(len(matrix))

        iterations = 0
        while np.any(negative):
            eigenvalues[negative] = 0
            matrix = eigenvectors.dot(np.diag(eigenvalues)).dot(eigenvectors.T)
            if iterations >= 10:
                break

            matrix = matrix - matrix * identity + identity

            max_value = np.abs(np.abs(matrix).max())
            if max_value > 1:
                matrix /= max_value

            eigenvalues, eigenvectors = scipy.linalg.eigh(matrix)
            negative = eigenvalues < 0
            iterations += 1

        return matrix

    @classmethod
    def _rebuild_correlation_matrix(cls, triangular_correlation):
        """Rebuild a valid correlation matrix from its lower half triangle.

        The input of this function is a list of lists of floats of size 1, 2, 3...n-1:

           [[c_{2,1}], [c_{3,1}, c_{3,2}], ..., [c_{n,1},...,c_{n,n-1}]]

        Corresponding to the values from the lower half of the original correlation matrix,
        **excluding** the diagonal.

        The output is the complete correlation matrix reconstructed using the given values
        and scaled to the :math:`[-1, 1]` range if necessary.

        Args:
            triangle_correlation (list[list[float]]):
                A list that contains lists of floats of size 1, 2, 3... up to ``n-1``,
                where ``n`` is the size of the target correlation matrix.

        Returns:
            numpy.ndarray:
                rebuilt correlation matrix.
        """
        zero = [0.0]
        size = len(triangular_correlation) + 1
        left = np.zeros((size, size))
        right = np.zeros((size, size))
        for idx, values in enumerate(triangular_correlation):
            values = values + zero * (size - idx - 1)
            left[idx + 1, :] = values
            right[:, idx + 1] = values

        correlation = left + right
        max_value = np.abs(correlation).max()
        if max_value > 1:
            correlation /= max_value

        correlation += np.identity(size)

        return cls._get_nearest_correlation_matrix(correlation).tolist()

    def _rebuild_gaussian_copula(self, model_parameters):
        """Rebuild the model params to recreate a Gaussian Multivariate instance.

        Args:
            model_parameters (dict):
                Sampled and reestructured model parameters.

        Returns:
            dict:
                Model parameters ready to recreate the model.
        """
        columns = []
        univariates = []
        for column, univariate in model_parameters["univariates"].items():
            columns.append(column)
            univariate["type"] = self.get_distribution_class(
                self._numerical_distributions.get(column, self.default_distribution)
            )
            if "scale" in univariate:
                univariate["scale"] = max(0, univariate["scale"])

            univariates.append(univariate)

        model_parameters["univariates"] = univariates
        model_parameters["columns"] = columns

        correlation = model_parameters.get("correlation")
        if correlation:
            model_parameters["correlation"] = self._rebuild_correlation_matrix(correlation)
        else:
            model_parameters["correlation"] = [[1.0]]

        return model_parameters

    def _get_likelihood(self, table_rows):
        return self._model.probability_density(table_rows)

    def _set_parameters(self, parameters):
        """Set copula model parameters.

        Args:
            dict:
                Copula flatten parameters.
        """
        parameters = unflatten_dict(parameters)
        if "num_rows" in parameters:
            num_rows = parameters.pop("num_rows")
            self._num_rows = 0 if pd.isna(num_rows) else max(0, int(round(num_rows)))

        if parameters:
            parameters = self._rebuild_gaussian_copula(parameters)
            self._model = multivariate.GaussianMultivariate.from_dict(parameters)



import warnings
from typing import Any, List, Optional

import numpy as np
import pandas as pd
import torch
from torch import optim
from torch.nn import (
    BatchNorm1d,
    Dropout,
    LeakyReLU,
    Linear,
    Module,
    ReLU,
    Sequential,
    functional,
)

# base 类已拆分，挪到 base.py
from sdgx.models.base import BaseSynthesizerModel
from sdgx.models.extension import hookimpl

# transformer 以及 sampler 已经拆分，挪到 transform/ 目录中
from sdgx.transform.sampler import DataSamplerCTGAN
from sdgx.transform.transformer import DataTransformerCTGAN

# 其他函数
from sdgx.utils.utils import random_state


class Discriminator(Module):
    """Discriminator for the CTGAN."""

    def __init__(self, input_dim, discriminator_dim, pac=10):
        super(Discriminator, self).__init__()
        dim = input_dim * pac
        self.pac = pac
        self.pacdim = dim
        seq = []
        for item in list(discriminator_dim):
            seq += [Linear(dim, item), LeakyReLU(0.2), Dropout(0.5)]
            dim = item

        seq += [Linear(dim, 1)]
        self.seq = Sequential(*seq)

    def calc_gradient_penalty(self, real_data, fake_data, device="cpu", pac=10, lambda_=10):
        """Compute the gradient penalty."""
        alpha = torch.rand(real_data.size(0) // pac, 1, 1, device=device)
        alpha = alpha.repeat(1, pac, real_data.size(1))
        alpha = alpha.view(-1, real_data.size(1))

        interpolates = alpha * real_data + ((1 - alpha) * fake_data)

        disc_interpolates = self(interpolates)

        gradients = torch.autograd.grad(
            outputs=disc_interpolates,
            inputs=interpolates,
            grad_outputs=torch.ones(disc_interpolates.size(), device=device),
            create_graph=True,
            retain_graph=True,
            only_inputs=True,
        )[0]

        gradients_view = gradients.view(-1, pac * real_data.size(1)).norm(2, dim=1) - 1
        gradient_penalty = ((gradients_view) ** 2).mean() * lambda_

        return gradient_penalty

    def forward(self, input_):
        """Apply the Discriminator to the `input_`."""
        assert input_.size()[0] % self.pac == 0
        return self.seq(input_.view(-1, self.pacdim))


class Residual(Module):
    """Residual layer for the CTGAN."""

    def __init__(self, i, o):
        super(Residual, self).__init__()
        self.fc = Linear(i, o)
        self.bn = BatchNorm1d(o)
        self.relu = ReLU()

    def forward(self, input_):
        """Apply the Residual layer to the `input_`."""
        out = self.fc(input_)
        out = self.bn(out)
        out = self.relu(out)
        return torch.cat([out, input_], dim=1)


class Generator(Module):
    """Generator for the CTGAN."""

    def __init__(self, embedding_dim, generator_dim, data_dim):
        super(Generator, self).__init__()
        dim = embedding_dim
        seq = []
        for item in list(generator_dim):
            seq += [Residual(dim, item)]
            dim += item
        seq.append(Linear(dim, data_dim))
        self.seq = Sequential(*seq)

    def forward(self, input_):
        """Apply the Generator to the `input_`."""
        data = self.seq(input_)
        return data


# 目前针对了较大的数据进行内存占用上的优化
# 总体思路是采用分批 load 进内存的方法，优化其占用空间的大小
class CTGAN(BaseSynthesizerModel):
    """Conditional Table GAN Synthesizer.

    This is the core class of the CTGAN project, where the different components
    are orchestrated together.
    For more details about the process, please check the [Modeling Tabular data using
    Conditional GAN](https://arxiv.org/abs/1907.00503) paper.

    Args:
        embedding_dim (int):
            Size of the random sample passed to the Generator. Defaults to 128.
        generator_dim (tuple or list of ints):
            Size of the output samples for each one of the Residuals. A Residual Layer
            will be created for each one of the values provided. Defaults to (256, 256).
        discriminator_dim (tuple or list of ints):
            Size of the output samples for each one of the Discriminator Layers. A Linear Layer
            will be created for each one of the values provided. Defaults to (256, 256).
        generator_lr (float):
            Learning rate for the generator. Defaults to 2e-4.
        generator_decay (float):
            Generator weight decay for the Adam Optimizer. Defaults to 1e-6.
        discriminator_lr (float):
            Learning rate for the discriminator. Defaults to 2e-4.
        discriminator_decay (float):
            Discriminator weight decay for the Adam Optimizer. Defaults to 1e-6.
        batch_size (int):
            Number of data samples to process in each step.
        discriminator_steps (int):
            Number of discriminator updates to do for each generator update.
            From the WGAN paper: https://arxiv.org/abs/1701.07875. WGAN paper
            default is 5. Default used is 1 to match original CTGAN implementation.
        log_frequency (boolean):
            Whether to use log frequency of categorical levels in conditional
            sampling. Defaults to ``True``.
        verbose (boolean):
            Whether to have print statements for progress results. Defaults to ``False``.
        epochs (int):
            Number of training epochs. Defaults to 300.
        pac (int):
            Number of samples to group together when applying the discriminator.
            Defaults to 10.
        cuda (bool):
            Whether to attempt to use cuda for GPU computation.
            If this is False or CUDA is not available, CPU will be used.
            Defaults to ``True``.
        memory_optimize(bool):
            设计上这个参数需要是 str 类型或者 bool 类型 可以是如下数值：
                - True：根据实际情况进行自动化内存优化，保证在能跑完的情况下，尽量多地利用内存，同时不报错
                - False：不采用任何内存优化策略
                - None：同 False
            该参数默认设为 False
    """

    def __init__(
        self,
        embedding_dim=128,
        generator_dim=(256, 256),
        discriminator_dim=(256, 256),
        generator_lr=2e-4,
        generator_decay=1e-6,
        discriminator_lr=2e-4,
        discriminator_decay=1e-6,
        batch_size=500,
        discriminator_steps=1,
        log_frequency=True,
        verbose=False,
        epochs=300,
        pac=10,
        cuda=True,
        memory_optimize=False,
    ):
        assert batch_size % 2 == 0

        self._embedding_dim = embedding_dim
        self._generator_dim = generator_dim
        self._discriminator_dim = discriminator_dim

        self._generator_lr = generator_lr
        self._generator_decay = generator_decay
        self._discriminator_lr = discriminator_lr
        self._discriminator_decay = discriminator_decay

        self._batch_size = batch_size
        self._discriminator_steps = discriminator_steps
        self._log_frequency = log_frequency
        self._verbose = verbose
        self._epochs = epochs
        self.pac = pac

        if not cuda or not torch.cuda.is_available():
            device = "cpu"
        elif isinstance(cuda, str):
            device = cuda
        else:
            device = "cuda"
        self._device = torch.device(device)

        self._generator = None

        # 是否启用内存优化，这里初步打算采用内存限制进行
        # 设计上这个参数需要是 str 类型或者 bool 类型 可以是如下数值：
        #   - True：根据实际情况进行自动化内存优化，保证在能跑完的情况下，尽量多地利用内存，同时不报错
        #   - False：不采用任何内存优化策略
        self.memory_optimize = memory_optimize

    @staticmethod
    def _gumbel_softmax(logits, tau=1, hard=False, eps=1e-10, dim=-1):
        """Deals with the instability of the gumbel_softmax for older versions of torch.

        For more details about the issue:
        https://drive.google.com/file/d/1AA5wPfZ1kquaRtVruCd6BiYZGcDeNxyP/view?usp=sharing

        Args:
            logits […, num_features]:
                Unnormalized log probabilities
            tau:
                Non-negative scalar temperature
            hard (bool):
                If True, the returned samples will be discretized as one-hot vectors,
                but will be differentiated as if it is the soft sample in autograd
            dim (int):
                A dimension along which softmax will be computed. Default: -1.

        Returns:
            Sampled tensor of same shape as logits from the Gumbel-Softmax distribution.
        """
        for _ in range(10):
            transformed = functional.gumbel_softmax(logits, tau=tau, hard=hard, eps=eps, dim=dim)
            if not torch.isnan(transformed).any():
                return transformed

        raise ValueError("gumbel_softmax returning NaN.")

    def _apply_activate(self, data):
        """Apply proper activation function to the output of the generator."""
        data_t = []
        st = 0
        for column_info in self._transformer.output_info_list:
            for span_info in column_info:
                if span_info.activation_fn == "tanh":
                    ed = st + span_info.dim
                    data_t.append(torch.tanh(data[:, st:ed]))
                    st = ed
                elif span_info.activation_fn == "softmax":
                    ed = st + span_info.dim
                    transformed = self._gumbel_softmax(data[:, st:ed], tau=0.2)
                    data_t.append(transformed)
                    st = ed
                else:
                    raise ValueError(f"Unexpected activation function {span_info.activation_fn}.")

        return torch.cat(data_t, dim=1)

    def _cond_loss(self, data, c, m):
        """Compute the cross entropy loss on the fixed discrete column."""
        loss = []
        st = 0
        st_c = 0
        for column_info in self._transformer.output_info_list:
            for span_info in column_info:
                if len(column_info) != 1 or span_info.activation_fn != "softmax":
                    # not discrete column
                    st += span_info.dim
                else:
                    ed = st + span_info.dim
                    ed_c = st_c + span_info.dim
                    tmp = functional.cross_entropy(
                        data[:, st:ed],
                        torch.argmax(c[:, st_c:ed_c], dim=1),
                        reduction="none",
                    )
                    loss.append(tmp)
                    st = ed
                    st_c = ed_c

        loss = torch.stack(loss, dim=1)  # noqa: PD013

        return (loss * m).sum() / data.size()[0]

    def _validate_discrete_columns(self, train_data, discrete_columns):
        """Check whether ``discrete_columns`` exists in ``train_data``.

        Args:
            train_data (numpy.ndarray or pandas.DataFrame):
                Training Data. It must be a 2-dimensional numpy array or a pandas.DataFrame.
            discrete_columns (list-like):
                List of discrete columns to be used to generate the Conditional
                Vector. If ``train_data`` is a Numpy array, this list should
                contain the integer indices of the columns. Otherwise, if it is
                a ``pandas.DataFrame``, this list should contain the column names.
        """
        if isinstance(train_data, pd.DataFrame):
            invalid_columns = set(discrete_columns) - set(train_data.columns)
        elif isinstance(train_data, np.ndarray):
            invalid_columns = []
            for column in discrete_columns:
                if column < 0 or column >= train_data.shape[1]:
                    invalid_columns.append(column)
        else:
            raise TypeError("``train_data`` should be either pd.DataFrame or np.array.")

        if invalid_columns:
            raise ValueError(f"Invalid columns found: {invalid_columns}")

    # OPTIMIZE 新增方法，完成在内存受限情况下的 CTGAN 训练
    @random_state
    def fit_optimze(self, train_data_iterator, discrete_columns: Optional[List] = [], epoches=10):
        """OPTIMIZE 新增方法，在内存受限情况下完成 CTGAN 训练

        参数列表：
            train_data_iterator (iterator)：
                train_data_iterator 是 带有训练数据的迭代器，
                需要根据 csv 文件，迭代返回训练数据，
                属于必填参数。
            discrete_columns (list-like)：
                描述离散特征的一个 python 列表。
                如果 训练数据的迭代器 中不含有列名，则返回列的编号，
                否则 ，返回列的名称（可以乱序），
                属于选填参数，默认为 python 空列表。
            epoches(int-like)：
                CTGAN 模型的迭代次数，
                选填参数，默认为10，但严肃应用中应必须填写。
        """
        if epochs is None:
            epochs = self._epochs

        pass

    @random_state
    def fit(
        self,
        train_data,
        discrete_columns: Optional[List] = None,
        train_data_iterator=None,
        epochs=None,
    ):
        """Fit the CTGAN Synthesizer models to the training data.

        Args:
            train_data (numpy.ndarray or pandas.DataFrame):
                Training Data. It must be a 2-dimensional numpy array or a pandas.DataFrame.
            train_data_iterator (iterator)：
                train_data_iterator 是 带有训练数据的迭代器
                需要根据 csv 文件，迭代返回训练数据
            discrete_columns (list-like):
                List of discrete columns to be used to generate the Conditional
                Vector. If ``train_data`` is a Numpy array, this list should
                contain the integer indices of the columns. Otherwise, if it is
                a ``pandas.DataFrame``, this list should contain the column names.
        """
        # set discrete_columns
        if not discrete_columns:
            discrete_columns = []

        # OPTIMIZE 检查 optimize 以及 train_data_iterator
        if self.memory_optimize and train_data_iterator is None:
            raise ValueError("train_data_iterator should not be None.")
        # 如果符合 optimize 的需求，则转到新实现的 optimize 方法，这样也不干扰老方法的顺利执行
        if self.memory_optimize and train_data_iterator is not None:
            self.fit_optimze(train_data_iterator, discrete_columns, epochs)
            return
        # OPTIMIZE 改动结束

        # 以下为原始的 fit 方法
        self._validate_discrete_columns(train_data, discrete_columns)

        # 参数检查
        if epochs is None:
            epochs = self._epochs
        else:
            warnings.warn(
                (
                    "`epochs` argument in `fit` method has been deprecated and will be removed "
                    "in a future version. Please pass `epochs` to the constructor instead"
                ),
                DeprecationWarning,
            )

        # 载入 transformer
        self._transformer = DataTransformerCTGAN()
        self._transformer.fit(train_data, discrete_columns)

        # 使用 transformer 处理数据
        train_data = self._transformer.transform(train_data)

        # 载入 sampler
        self._data_sampler = DataSamplerCTGAN(
            train_data, self._transformer.output_info_list, self._log_frequency
        )

        # data dim 从 transformer 中取得
        data_dim = self._transformer.output_dimensions

        # sampler 作为参数给到 Generator 以及 Discriminator
        self._generator = Generator(
            self._embedding_dim + self._data_sampler.dim_cond_vec(),
            self._generator_dim,
            data_dim,
        ).to(self._device)

        discriminator = Discriminator(
            data_dim + self._data_sampler.dim_cond_vec(),
            self._discriminator_dim,
            pac=self.pac,
        ).to(self._device)

        # 初始化 optimizer G 以及 D
        optimizerG = optim.Adam(
            self._generator.parameters(),
            lr=self._generator_lr,
            betas=(0.5, 0.9),
            weight_decay=self._generator_decay,
        )

        optimizerD = optim.Adam(
            discriminator.parameters(),
            lr=self._discriminator_lr,
            betas=(0.5, 0.9),
            weight_decay=self._discriminator_decay,
        )

        mean = torch.zeros(self._batch_size, self._embedding_dim, device=self._device)
        std = mean + 1

        # 开始执行 fit  流程，直到结束
        steps_per_epoch = max(len(train_data) // self._batch_size, 1)
        for i in range(epochs):
            for id_ in range(steps_per_epoch):
                for n in range(self._discriminator_steps):
                    fakez = torch.normal(mean=mean, std=std)

                    condvec = self._data_sampler.sample_condvec(self._batch_size)
                    if condvec is None:
                        c1, m1, col, opt = None, None, None, None
                        real = self._data_sampler.sample_data(self._batch_size, col, opt)
                    else:
                        c1, m1, col, opt = condvec
                        c1 = torch.from_numpy(c1).to(self._device)
                        m1 = torch.from_numpy(m1).to(self._device)
                        fakez = torch.cat([fakez, c1], dim=1)

                        perm = np.arange(self._batch_size)
                        np.random.shuffle(perm)
                        real = self._data_sampler.sample_data(
                            self._batch_size, col[perm], opt[perm]
                        )
                        c2 = c1[perm]

                    fake = self._generator(fakez)
                    fakeact = self._apply_activate(fake)

                    real = torch.from_numpy(real.astype("float32")).to(self._device)

                    if c1 is not None:
                        fake_cat = torch.cat([fakeact, c1], dim=1)
                        real_cat = torch.cat([real, c2], dim=1)
                    else:
                        real_cat = real
                        fake_cat = fakeact

                    y_fake = discriminator(fake_cat)
                    y_real = discriminator(real_cat)

                    pen = discriminator.calc_gradient_penalty(
                        real_cat, fake_cat, self._device, self.pac
                    )
                    loss_d = -(torch.mean(y_real) - torch.mean(y_fake))

                    optimizerD.zero_grad(set_to_none=False)
                    pen.backward(retain_graph=True)
                    loss_d.backward()
                    optimizerD.step()

                fakez = torch.normal(mean=mean, std=std)
                condvec = self._data_sampler.sample_condvec(self._batch_size)

                if condvec is None:
                    c1, m1, col, opt = None, None, None, None
                else:
                    c1, m1, col, opt = condvec
                    c1 = torch.from_numpy(c1).to(self._device)
                    m1 = torch.from_numpy(m1).to(self._device)
                    fakez = torch.cat([fakez, c1], dim=1)

                fake = self._generator(fakez)
                fakeact = self._apply_activate(fake)

                if c1 is not None:
                    y_fake = discriminator(torch.cat([fakeact, c1], dim=1))
                else:
                    y_fake = discriminator(fakeact)

                if condvec is None:
                    cross_entropy = 0
                else:
                    cross_entropy = self._cond_loss(fake, c1, m1)

                loss_g = -torch.mean(y_fake) + cross_entropy

                optimizerG.zero_grad(set_to_none=False)
                loss_g.backward()
                optimizerG.step()

            if self._verbose:
                print(
                    f"Epoch {i+1}, Loss G: {loss_g.detach().cpu(): .4f},"  # noqa: T001
                    f"Loss D: {loss_d.detach().cpu(): .4f}",
                    flush=True,
                )

    @random_state
    def sample(self, n, condition_column=None, condition_value=None):
        """Sample data similar to the training data.

        Choosing a condition_column and condition_value will increase the probability of the
        discrete condition_value happening in the condition_column.

        Args:
            n (int):
                Number of rows to sample.
            condition_column (string):
                Name of a discrete column.
            condition_value (string):
                Name of the category in the condition_column which we wish to increase the
                probability of happening.

        Returns:
            numpy.ndarray or pandas.DataFrame
        """
        if condition_column is not None and condition_value is not None:
            condition_info = self._transformer.convert_column_name_value_to_id(
                condition_column, condition_value
            )
            global_condition_vec = self._data_sampler.generate_cond_from_condition_column_info(
                condition_info, self._batch_size
            )
        else:
            global_condition_vec = None

        steps = n // self._batch_size + 1
        data = []
        for i in range(steps):
            mean = torch.zeros(self._batch_size, self._embedding_dim)
            std = mean + 1
            fakez = torch.normal(mean=mean, std=std).to(self._device)

            if global_condition_vec is not None:
                condvec = global_condition_vec.copy()
            else:
                condvec = self._data_sampler.sample_original_condvec(self._batch_size)

            if condvec is None:
                pass
            else:
                c1 = condvec
                c1 = torch.from_numpy(c1).to(self._device)
                fakez = torch.cat([fakez, c1], dim=1)

            fake = self._generator(fakez)
            fakeact = self._apply_activate(fake)
            data.append(fakeact.detach().cpu().numpy())

        data = np.concatenate(data, axis=0)
        data = data[:n]

        return self._transformer.inverse_transform(data)


@hookimpl
def register(manager):
    manager.register("CTGAN", CTGAN)



# TBD
# 主要用于存放 sdg 中特有的的报错信息


class SdgxError(Exception):
    """Base class for exceptions in this module."""

    pass


class NonParametricError(Exception):
    """Exception to indicate that a model is not parametric."""


class ModelNotFoundError(SdgxError):
    pass


class ModelRegisterError(SdgxError):
    pass


class ModelInitializationError(SdgxError):
    pass



# 本例用于测试 transformer 对象的正常使用

import os
import shutil
import sys

import pytest

_HERE = os.path.dirname(__file__)
sys.path.append(os.getcwd())

from sdgx.transform.transformer import DataTransformerCTGAN
from sdgx.transform.transformer_opt import DataTransformer
from sdgx.utils.io.csv_utils import *


def test_transformer_original():
    demo_data, discrete_cols = get_demo_single_table()
    ctgan_transformer = DataTransformerCTGAN()
    ctgan_transformer.fit(demo_data, discrete_cols)
    transformed_data = ctgan_transformer.transform(demo_data)


def test_transformer_opt():
    # 测试经过内存优化之后的 transformer
    demo_data_path = "./dataset/adult.csv"
    _, discrete_cols = get_demo_single_table()
    opt_transformer = DataTransformer()
    opt_transformer.fit(demo_data_path, discrete_cols)
    opt_transformer.transform(demo_data_path, "./output_tmp.csv")
    # inverse transforme
    x = pd.read_csv("./output_tmp.csv")
    x = x.to_numpy()
    inverse_x = opt_transformer.inverse_transform(x)
    inverse_x = opt_transformer.inverse_transform(
        x, output_path="inverse_tmp.csv", write_header=True
    )

    # remove tmp files
    shutil.os.remove("inverse_tmp.csv")
    shutil.os.remove("output_tmp.csv")
    pass


if __name__ == "__main__":
    pytest.main(["-vv", "-s", __file__])



"""
    DataTransformer 模块：

    目前使用了CTGAN开源项目中的代码
    后续还会根据实际业务需求进一步进行改写
    以及进行一些性能优化
"""

from collections import namedtuple
from typing import List, Optional

import numpy as np
import pandas as pd
from joblib import Parallel, delayed
from rdt.transformers import ClusterBasedNormalizer, OneHotEncoder

SpanInfo = namedtuple("SpanInfo", ["dim", "activation_fn"])
ColumnTransformInfo = namedtuple(
    "ColumnTransformInfo",
    ["column_name", "column_type", "transform", "output_info", "output_dimensions"],
)


class DataTransformerCTGAN(object):
    """Data Transformer.

    Model continuous columns with a BayesianGMM and normalized to a scalar [0, 1] and a vector.
    Discrete columns are encoded using a scikit-learn OneHotEncoder.
    """

    def __init__(self, max_clusters=10, weight_threshold=0.005):
        """Create a data transformer.

        Args:
            max_clusters (int):
                Maximum number of Gaussian distributions in Bayesian GMM.
            weight_threshold (float):
                Weight threshold for a Gaussian distribution to be kept.
        """
        self._max_clusters = max_clusters
        self._weight_threshold = weight_threshold

    def _fit_continuous(self, data):
        """Train Bayesian GMM for continuous columns.

        Args:
            data (pd.DataFrame):
                A dataframe containing a column.

        Returns:
            namedtuple:
                A ``ColumnTransformInfo`` object.
        """
        column_name = data.columns[0]

        # 这里进行了修改，用于防止未来的兼容性问题，同时避免出现以下警告
        # ======================================================
        # FutureWarning: Future versions of RDT
        # will not support the 'model_missing_values' parameter.
        # Please switch to using the 'missing_value_generation'
        # parameter to select your strategy.
        # ======================================================

        # 修改前
        # gm = ClusterBasedNormalizer(model_missing_values=True, max_clusters=min(len(data), 10))
        # 修改后
        gm = ClusterBasedNormalizer(
            missing_value_generation="from_column", max_clusters=min(len(data), 10)
        )
        gm.fit(data, column_name)
        num_components = sum(gm.valid_component_indicator)

        return ColumnTransformInfo(
            column_name=column_name,
            column_type="continuous",
            transform=gm,
            output_info=[SpanInfo(1, "tanh"), SpanInfo(num_components, "softmax")],
            output_dimensions=1 + num_components,
        )

    def _fit_discrete(self, data):
        """Fit one hot encoder for discrete column.

        Args:
            data (pd.DataFrame):
                A dataframe containing a column.

        Returns:
            namedtuple:
                A ``ColumnTransformInfo`` object.
        """
        column_name = data.columns[0]
        ohe = OneHotEncoder()
        ohe.fit(data, column_name)
        num_categories = len(ohe.dummies)

        return ColumnTransformInfo(
            column_name=column_name,
            column_type="discrete",
            transform=ohe,
            output_info=[SpanInfo(num_categories, "softmax")],
            output_dimensions=num_categories,
        )

    def fit(self, raw_data, discrete_columns: Optional[List] = None):
        """Fit the ``DataTransformer``.

        Fits a ``ClusterBasedNormalizer`` for continuous columns and a
        ``OneHotEncoder`` for discrete columns.

        This step also counts the #columns in matrix data and span information.
        """
        self.output_info_list = []
        self.output_dimensions = 0
        self.dataframe = True
        if not discrete_columns:
            discrete_columns = []

        if not isinstance(raw_data, pd.DataFrame):
            self.dataframe = False
            # work around for RDT issue #328 Fitting with numerical column names fails
            discrete_columns = [str(column) for column in discrete_columns]
            column_names = [str(num) for num in range(raw_data.shape[1])]
            raw_data = pd.DataFrame(raw_data, columns=column_names)

        self._column_raw_dtypes = raw_data.infer_objects().dtypes
        self._column_transform_info_list = []
        for column_name in raw_data.columns:
            if column_name in discrete_columns:
                column_transform_info = self._fit_discrete(raw_data[[column_name]])
            else:
                column_transform_info = self._fit_continuous(raw_data[[column_name]])

            self.output_info_list.append(column_transform_info.output_info)
            self.output_dimensions += column_transform_info.output_dimensions
            self._column_transform_info_list.append(column_transform_info)

    def _transform_continuous(self, column_transform_info, data):
        column_name = data.columns[0]
        flattened_column = data[column_name].to_numpy().flatten()
        data = data.assign(**{column_name: flattened_column})
        gm = column_transform_info.transform
        transformed = gm.transform(data)

        #  Converts the transformed data to the appropriate output format.
        #  The first column (ending in '.normalized') stays the same,
        #  but the lable encoded column (ending in '.component') is one hot encoded.
        output = np.zeros((len(transformed), column_transform_info.output_dimensions))
        output[:, 0] = transformed[f"{column_name}.normalized"].to_numpy()
        index = transformed[f"{column_name}.component"].to_numpy().astype(int)
        output[np.arange(index.size), index + 1] = 1.0

        return output

    def _transform_discrete(self, column_transform_info, data):
        ohe = column_transform_info.transform
        return ohe.transform(data).to_numpy()

    def _synchronous_transform(self, raw_data, column_transform_info_list):
        """Take a Pandas DataFrame and transform columns synchronous.

        Outputs a list with Numpy arrays.
        """
        column_data_list = []
        for column_transform_info in column_transform_info_list:
            column_name = column_transform_info.column_name
            data = raw_data[[column_name]]
            if column_transform_info.column_type == "continuous":
                column_data_list.append(self._transform_continuous(column_transform_info, data))
            else:
                column_data_list.append(self._transform_discrete(column_transform_info, data))

        return column_data_list

    def _parallel_transform(self, raw_data, column_transform_info_list):
        """Take a Pandas DataFrame and transform columns in parallel.

        Outputs a list with Numpy arrays.
        """
        processes = []
        for column_transform_info in column_transform_info_list:
            column_name = column_transform_info.column_name
            data = raw_data[[column_name]]
            process = None
            if column_transform_info.column_type == "continuous":
                process = delayed(self._transform_continuous)(column_transform_info, data)
            else:
                process = delayed(self._transform_discrete)(column_transform_info, data)
            processes.append(process)

        return Parallel(n_jobs=-1)(processes)

    def transform(self, raw_data):
        """Take raw data and output a matrix data."""
        if not isinstance(raw_data, pd.DataFrame):
            column_names = [str(num) for num in range(raw_data.shape[1])]
            raw_data = pd.DataFrame(raw_data, columns=column_names)

        # Only use parallelization with larger data sizes.
        # Otherwise, the transformation will be slower.
        if raw_data.shape[0] < 500:
            column_data_list = self._synchronous_transform(
                raw_data, self._column_transform_info_list
            )
        else:
            column_data_list = self._parallel_transform(raw_data, self._column_transform_info_list)

        return np.concatenate(column_data_list, axis=1).astype(float)

    def _inverse_transform_continuous(self, column_transform_info, column_data, sigmas, st):
        gm = column_transform_info.transform
        data = pd.DataFrame(column_data[:, :2], columns=list(gm.get_output_sdtypes()))
        data[data.columns[1]] = np.argmax(column_data[:, 1:], axis=1)
        if sigmas is not None:
            selected_normalized_value = np.random.normal(data.iloc[:, 0], sigmas[st])
            data.iloc[:, 0] = selected_normalized_value

        return gm.reverse_transform(data)

    def _inverse_transform_discrete(self, column_transform_info, column_data):
        ohe = column_transform_info.transform
        data = pd.DataFrame(column_data, columns=list(ohe.get_output_sdtypes()))
        return ohe.reverse_transform(data)[column_transform_info.column_name]

    def inverse_transform(self, data, sigmas=None):
        """Take matrix data and output raw data.

        Output uses the same type as input to the transform function.
        Either np array or pd dataframe.
        """
        st = 0
        recovered_column_data_list = []
        column_names = []
        for column_transform_info in self._column_transform_info_list:
            dim = column_transform_info.output_dimensions
            column_data = data[:, st : st + dim]
            if column_transform_info.column_type == "continuous":
                recovered_column_data = self._inverse_transform_continuous(
                    column_transform_info, column_data, sigmas, st
                )
            else:
                recovered_column_data = self._inverse_transform_discrete(
                    column_transform_info, column_data
                )

            recovered_column_data_list.append(recovered_column_data)
            column_names.append(column_transform_info.column_name)
            st += dim

        recovered_data = np.column_stack(recovered_column_data_list)
        recovered_data = pd.DataFrame(recovered_data, columns=column_names).astype(
            self._column_raw_dtypes
        )
        if not self.dataframe:
            recovered_data = recovered_data.to_numpy()

        return recovered_data

    def convert_column_name_value_to_id(self, column_name, value):
        """Get the ids of the given `column_name`."""
        discrete_counter = 0
        column_id = 0
        for column_transform_info in self._column_transform_info_list:
            if column_transform_info.column_name == column_name:
                break
            if column_transform_info.column_type == "discrete":
                discrete_counter += 1

            column_id += 1

        else:
            raise ValueError(f"The column_name `{column_name}` doesn't exist in the data.")

        ohe = column_transform_info.transform
        data = pd.DataFrame([value], columns=[column_transform_info.column_name])
        one_hot = ohe.transform(data).to_numpy()[0]
        if sum(one_hot) == 0:
            raise ValueError(f"The value `{value}` doesn't exist in the column `{column_name}`.")

        return {
            "discrete_column_id": discrete_counter,
            "column_id": column_id,
            "value_id": np.argmax(one_hot),
        }



import urllib.request
from pathlib import Path

import pandas as pd


def get_demo_single_table():
    """ """
    demo_data_path = Path("./dataset/adult.csv")
    if not demo_data_path.exists():
        # Download from datahub
        demo_data_path.parent.mkdir(parents=True, exist_ok=True)

        # FIXME: Use logging
        print("Downloading demo data from datahub.io to ./dataset/adult.csv")
        url = "https://datahub.io/machine-learning/adult/r/adult.csv"
        urllib.request.urlretrieve(url, demo_data_path)

    pd_obj = pd.read_csv(demo_data_path)
    discrete_cols = [
        "workclass",
        "education",
        "marital-status",
        "occupation",
        "relationship",
        "race",
        "sex",
        "native-country",
        "class",
    ]
    return pd_obj, discrete_cols


# 获取csv形式的单表数据
def get_single_table(input_path):
    pass


def get_csv_column(input_path, column_name):
    """按 column 形式读取 csv，返回 pd.DataFrame 的单列数据

    输入参数:
        input_path (str):
            作为输入的 csv 路径
        column_name (str)：
            需要提取的 csv 列名

    返回对象说明:
        namedtuple对象 (pd.DataFrame):
            返回单个 ``pd.DataFrame`` 对象
    """
    df_col = pd.read_csv(input_path, usecols=[column_name])
    return df_col


def get_csv_column_count(input_path, id_col=False):
    """get the column number from csv, the input is the path

    Args:
        input_path (str):
            the path of csv file
        id_col (bool):
            whether csv file contains id column,
            if true, the count of column will minus by 1

    Returns:
        cnt (int):
            the count of column in csv
    """
    first_line = None
    with open(input_path) as f:
        first_line = f.readline()
    cnt = len(first_line.split(","))
    if id_col:
        cnt = cnt - 1
    return cnt
